/*************************************************************************

FILE:	lex.c

This file contains the code of the lexical analyzer, able to read the
input files codified according to a ".mc", ".rg", or ".prb" format.
It is used by all the programs that need to read one of these files.

FUNCTIONS CONTAINED:
	lex()
	initlex()
	E_symlong()
	E_illchar()
	E_unknown()
	E_specexp()
	E_valerr()
	E_charexp()
*/

# define LEXMAIN

# include	<stdio.h>
# include	<stdlib.h>
# include	"port.h"
# include	"lex.h"

extern	double	pow();


int E_illchar(ch)
int	ch;
{
	fprintf(stderr,"ERROR line %d, illegal character (%o)\n",Line,ch);
	exit(ER_CODE);
}


int E_symlong() {
	fprintf(stderr,"ERROR line %d, symbol too long\n",Line);
	exit(ER_CODE);
}


int E_unknown(st) char *st;
{
	fprintf(stderr,"ERROR line %d, unknown or unexpected keyword (%s)\n",Line,st);
	exit(ER_CODE);
}


int E_specexp(st) char	*st; {
	fprintf(stderr,"ERROR line %d, %s specification expected\n",Line,st);
	exit(ER_CODE);
} /* end of E_specexp() */


int E_valerr(st) char	*st; {
	fprintf(stderr,"ERROR line %d, %s value out of range\n",Line,st);
	exit(ER_CODE);
} /* end of E_inderr() */


int E_charexp(cc) int cc; {
	fprintf(stderr,"ERROR line %d, '%c' expected\n",Line,cc);
	exit(ER_CODE);
}


void initlex()

/*************************************************************************
*/

{
	int	i;
static	char	auxGr[ALPHA_SIZE];

	/* initialize the variables needed by lexical analyzer */
	Gr = &auxGr[1];
	/* Gr is indexed from -1 to ALPHA_SIZE-2 */
	for (i = ALPHA_SIZE-1; i >= -1; i--)
		Gr[i] = C_ILLE;
	Gr[EOF] = C_EOF;
	for (i = 0; i < 10; i++)
		Gr[DIGITS[i]] = C_DIGI;
	for (i = 0; i < 53; i++)
		Gr[LETTERS[i]] = C_LETT;
	Gr['#'] = C_REMA;
	Gr[';'] = C_SEMI;
	Gr[':'] = C_COLO;
	Gr['='] = C_EQUA;
	Gr['.'] = C_PERI;
	Gr['+'] = C_PLUS;
	Gr['-'] = C_MINU;
	Gr[' '] = C_SPAC;
	Gr['\t'] = C_SPAC;
	Gr['\n'] = C_NEWL;
} /* end of initlex() */

int	lex(ff)
FILE	*ff;			/* the input file */

/*************************************************************************
Lexical analyzer, returns a token, and, possibly, a value in Vdouble.
It assumes that "Chrd" already contains the first caracter not yet
consumed (possibly EOF).

RETURNS:	a lexical token (integer corresponding to a "# define").
*/

{
	int	i,j;
	int	position;		/* free position in Buffer */
	int	sign;			/* the sign of the number */
	int	sexpo;			/* the sign of the exponent */
	int	dpow;			/* number of decimal positions */
	double	expo;			/* the value of the exponent */

for (;;) {
	switch (Gr[Chrd]) {
	case C_LETT:
		Buffer[(position = 0)] = Chrd;
		while (Gr[(Chrd = getc(ff))] == C_LETT || Gr[Chrd] == C_DIGI ||
		Gr[Chrd] == C_PERI) {
			if (++position == MAXBUFFER)
				E_symlong();
			Buffer[position] = Chrd;
		}
		Buffer[position+1] = '\0';
		for (i = 0; i < Nkeyword; i++)
			for (j = 0; Table[i][j] == Buffer[j]; j++)
				if (Buffer[j] == '\0')
					return(i);
		return(L_STRING);
	case C_MINU:
		Chrd = getc(ff);
		sign = -1;
		goto SKIP1;
	case C_PLUS:
		Chrd = getc(ff);
	case C_DIGI:
		sign = 1;
SKIP1:		Vdouble = 0.0;
		while (Gr[Chrd] == C_DIGI) {
			Vdouble = Vdouble * 10.0 + Chrd - '0';
			Chrd = getc(ff);
		}
		if (Gr[Chrd] == C_PERI) {
			goto DECIMALPOINT;
		} else if (Chrd == 'e') {
			dpow = 0;
			goto EXPONOTATION;
		} else {
			Vdouble = Vdouble * sign;
			return(L_NUMBER);
		}
	case C_PERI:
		sign = 1;
		Vdouble = 0.0;
DECIMALPOINT:	for (dpow = 0; Gr[(Chrd = getc(ff))] == C_DIGI; dpow++)
			Vdouble = Vdouble * 10.0 + Chrd - '0';
		if (Chrd != 'e') {
			Vdouble *= sign * pow((double)(10.0),(double)(-dpow));
			return(L_FLOAT);
		}
EXPONOTATION:	switch (Gr[Chrd = getc(ff)]) {
		case C_MINU:
			 Chrd = getc(ff);
			 sexpo = -1;
			 break;
		case C_PLUS:
			 Chrd = getc(ff);
		case C_DIGI:
			 sexpo = 1;
			 break;
		default:
			 E_illchar(Chrd);
		}
		for (expo = 0; Gr[Chrd] == C_DIGI; ) {
			 expo = expo * 10 + Chrd - '0';
			 Chrd = getc(ff);
		}
		Vdouble *= sign * pow(10.0,(sexpo * expo - dpow));
		return(L_FLOAT);
	case C_COLO:
		Chrd = getc(ff);
		return(L_COLON);
	case C_EQUA:
		Chrd = getc(ff);
		return(L_EQUAL);
	case C_SPAC:
		Chrd = getc(ff);
		break;
	case C_NEWL:
		Line++;
		Chrd = getc(ff);
		break;
	case C_SEMI:
		Chrd = getc(ff);
		return(L_SEMICOLON);
	case C_REMA:
		while ((Chrd = getc(ff)) != NEWLINE);
		Line++;
		Chrd = getc(ff);
		break;
	case C_EOF:
		return(L_EOF);
	case C_ILLE:
		E_illchar(Chrd);
	} /* switch */
} /* for (;;) */
} /* end of lex() */


