
/*************************************************************************

FILE:	reach.c

It generates the reachability graph and the coresponding matrix.

FUNCTIONS CONTAINED:
	cmp_mark()
	search_ins()
	normalize_probs()
	reach_graph()
	graph_analysis()
	AHUsearch()
	create_matrix()
*/

# include	<stdio.h>
# include	<stdlib.h>
# include	"port.h"
# include	"const.h"
# include	"type.h"
# include	"var.h"
# include	"utility.h"
# include	"cspl.h"
# include	"print.h"
# include	"net.h"

int	cmp_mark(m1,m2)
int	*m1;	/* array containing the first marking */
ME	*m2;	/* ME containing the second marking */

/*************************************************************************
It compares the marking "m1" with the marking "m2".
RETURNS:
CMP_EQ, CMP_GT, CMP_LT according to the relative position of "m1" and "m2"
in lexicographic order:
1(2) 3(4)	<	1(3) 3(4)
1(2) 3(4)	<	1(2) 4(7)
1(2) 3(4)	<	1(2) 3(4) 4(7)
*/

{
	int	i,diff;

	for (i = 0; i < Nplace; i++) {
		diff =  m1[i] - m2->mk[i];
		if (diff != 0)
			return((diff < 0) ? CMP_LT : CMP_GT);
	}
	return(CMP_EQ);
} /* cmp_mark() */

ME	*search_ins(plast,pmark)
ME	**plast;	/* last marking in canonic order */
int	*pmark;		/* marking to be searched */

/*************************************************************************
It searches the AVL-tree to find marking "pmark".
It allocates a new ME if not found.
RETURNS:
the pointer to the ME of the searched marking in the AVL-tree.
*/

{
	ME	*r,*p,*q,*s,*t;
	int	i;
	int	comp,scomp,invscomp;

	p = s = Proot;
	t = NULL;
	/* search the node in the AVL tree */
	for (;;) {
		comp = cmp_mark(pmark,p);
		if (comp != CMP_EQ) {	/* move left or right */
			if ((q = p->plr[comp]) != NULL) {
				if (q->bal != CMP_EQ) {	/* adjust t and s */
					t = p;
					s = q;
				}
				p = q;
			} else
				break;		/* marking not already present */
		} else
			return(p);	/* marking already present */
	}	/* end of search */
	/* insert marking */
	q = get_ME((unsigned) Nplace);	/* get a node */
	p->plr[comp] = q;		/* connect it */
	(*plast)->pcan = q;
	(*plast) = q;
	q->pcan = q->plr[CMP_LT] = q->plr[CMP_GT] = NULL;	/* fill it */
	q->bal = CMP_EQ;
	for (i = 0; i < Nplace; i++) 
		q->mk[i] = pmark[i];
	q->index = 0;
	q->prow = NULL;

/* now the AVL tree may be out of balance: check */
	/* adjustment of balance between S and Q nodes */
	scomp = cmp_mark(pmark,s);
	r = s->plr[scomp];
	p = r;
	while (p != q) {
		comp = cmp_mark(pmark,p);
		p->bal = comp;
		p = p->plr[comp];
	}
	/* balancing act */
	if (s->bal == CMP_EQ)
		s->bal = scomp;	 /* the tree is now higher */
	else if (s->bal != scomp)
		s->bal = CMP_EQ;/* the tree is now more balanced */
	else {			/* the tree is now out of balance */
		invscomp = (scomp == CMP_GT) ? CMP_LT : CMP_GT;
		if (r->bal == scomp) {	/* single rotation is needed */
			p = r;
			s->plr[scomp] = r->plr[invscomp];
			r->plr[invscomp] = s;
			s->bal = r->bal = CMP_EQ;
		} else {		/* double rotation is needed */
			p = r->plr[invscomp];
			r->plr[invscomp] = p->plr[scomp];
			p->plr[scomp] = r;
			s->plr[scomp] = p->plr[invscomp];
			p->plr[invscomp] = s;
			if (p->bal == scomp) {
				s->bal = invscomp;
				r->bal = CMP_EQ;
			} else if (p->bal == invscomp) {
				s->bal = CMP_EQ;
				r->bal = scomp;
			} else
				s->bal = r->bal = CMP_EQ;
			p->bal = CMP_EQ;
		} /* end of double rotation */
		/* finishing touch */
		if (t != NULL)
			t->plr[(s == t->plr[CMP_GT]) ? CMP_GT : CMP_LT] = p;
		else
			Proot = p;
	} /* end of the tree is now out of balance */
	return(q);
} /* search_ins() */

void	put_ret_m0_entries()

/*************************************************************************
It puts entries from absorbing markings M0 
*/

{
	EE	*auxee;
	ME	*auxme;

	for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan) {
		if (auxme->type == MAR_ABS) {
			auxee = auxme->prow = get_EE();
			auxee->tr = MCT_ABS;
			auxee->value = FOpt[FOP_ABS_RET_M0].val;
			auxee->col.mk = Pcan;
			auxee->row.pt = NULL;
		}
	}
}

void	normalize_probs()

/*************************************************************************
It normalizes the probabilities on each row.
Selfentries are eliminated by setting their value to 0, they will be released
in "create_matrix".
*/

{
	double	sum;
	double	dsum;
	EE	*auxee;
	ME	*auxme;

	if (Nvan == 0)
		return;
	for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan) {
		if (auxme->type != MAR_VAN && auxme->type != MAR_VLO)
			continue;
		sum = 0;
		dsum = 0;
		for (auxee = auxme->prow; auxee != NULL; auxee = auxee->row.pt)
			if (auxee->col.mk == auxme)/* check for self entry */
			{
				auxee->value = 0.0;
				auxee->d_value = 0.0;
			}
			else
			{
				sum += auxee->value;
				dsum += auxee->d_value;
			}
		sum = 1.0 / sum;
		for (auxee = auxme->prow; auxee != NULL; auxee = auxee->row.pt)
		{
			auxee->value *= sum;
			auxee->d_value =
				(auxee->d_value + auxee->value * dsum) * sum;
		}
	}
} /* normalize_probs() */

int	reach_graph()

/*************************************************************************
It generates the reachability set, and the reachability graph.
The marking canonic order reflects a breadth-first search.
Absorbing markings are managed.
RETURNS:
RES_ERROR or RES_NOERR;
*/

{
extern	char	*pr_place();
extern	char	*pr_trans();
extern	void	graph_analysis();
extern	ME	*Currmark;
	char	aux1[30],aux2[30];
	int	*newmark;
	ME	*auxme,*last;
	EE	*auxee;
	int	i,prio,input,output;
	TR	*maxpriotrans,*auxtr;
	double	entrval;
	double	d_entrval;
	int	errflg = RES_NOERR;

	Narcs = Ntan = Nabs = Nvan = 0;

	{ int minprio,count,oldprio;
		/* order the transitions according to priority */
		maxpriotrans = NULL;
		oldprio = -1;
		count = 0;
		while (count < Ntrans) {
			/* find min priority */
			minprio = MAX_PRIO + 1;
			for (i = 0; i < Ntrans;  i++)
				if (Arrtrans[i].prio < minprio && Arrtrans[i].prio > oldprio)
					minprio = Arrtrans[i].prio;
			oldprio = minprio;
			/* insert the transitions with priority "minprio" */
			for (i = Ntrans; --i >= 0; )
				if (Arrtrans[i].prio == minprio) {
					count++;
					Arrtrans[i].pnext = maxpriotrans;
					maxpriotrans = &Arrtrans[i];
				}
		}
	}
	/* allocate the memory for "newmark" */
	newmark = (int *)(new_mem((unsigned int)(sizeof(int) * Nplace)));
	/* insert the initial marking in the AVL-tree */
	last = Proot = Pcan = Currmark = get_ME((unsigned) Nplace);
	for (i = 0; i < Nplace; i++)
		Currmark->mk[i] = Arrplace[i].initial;
	Currmark->bal = CMP_EQ;
	Currmark->prow = NULL;
	Currmark->pcan = Currmark->plr[CMP_LT] = Currmark->plr[CMP_GT] = NULL;

	/* while new markings are found, go on */
	for (; Currmark != NULL; Currmark = Currmark->pcan) {
		/* check whether Currmark is legal */
		if (assert() == RES_ERROR) {
			fprintf(stderr,"Assertion not satisfied in marking:\n");
			pr_mark(Currmark,stderr,USE_NAME);
			errflg = RES_ERROR;
			goto RETLBL;
		}
		if (IOpt[IOP_DEBUG].val == VAL_YES) {
			Currmark->index = 0;
			Currmark->type = MAR_NOVAL;
			pr_mark(Currmark,stderr,USE_NAME);
		}	
		/* consider the enabled transitions */
		Currmark->type = MAR_ABS;
		for (auxtr = maxpriotrans; auxtr != NULL; auxtr = auxtr->pnext) {
			if (Currmark->type == MAR_ABS) {
				/* no enabled transition found yet */
				if (!internal_enabled(auxtr))
					continue;
				prio = auxtr->prio;
				/* compute the rate/prob */
				switch (auxtr->dep) {
				case DEP_NO:
					entrval = auxtr->val;
					d_entrval = auxtr->dval;
					break;
				case DEP_PL:
					entrval = auxtr->val * Currmark->mk[auxtr->pl];
					d_entrval = auxtr->dval * Currmark->mk[auxtr->pl];
					break;
				case DEP_MK:
					entrval = (*auxtr->fun)(auxtr->i1,auxtr->i2);
					d_entrval = (*auxtr->dfun)(auxtr->i1,auxtr->i2);
					break;
				}
				switch (auxtr->type) {
				case DIS_IMM:
					Currmark->type = MAR_VAN;
					break;
				case DIS_EXP:
					Currmark->type = MAR_TAN;
					break;
				}
			} else {
				if (auxtr->prio < prio)
					break;
				if (!internal_enabled(auxtr))
					continue;
				/* compute the rate/prob */
				switch (auxtr->dep) {
				case DEP_NO:
					entrval = auxtr->val;
					d_entrval = auxtr->dval;
					break;
				case DEP_PL:
					entrval = auxtr->val * Currmark->mk[auxtr->pl];
					d_entrval = auxtr->dval * Currmark->mk[auxtr->pl];
					break;
				case DEP_MK:
					entrval = (*auxtr->fun)(auxtr->i1,auxtr->i2);
					d_entrval = (*auxtr->dfun)(auxtr->i1,auxtr->i2);
					break;
				}
				switch (auxtr->type) {
				case DIS_IMM:
					if (Currmark->type != MAR_VAN) {
						fprintf(stderr,"ERROR: transitions %s and %s enabled in marking:\n",pr_trans(aux1,auxtr),pr_trans(aux2,&Arrtrans[Currmark->prow->tr]));
						pr_mark(Currmark,stderr,USE_NAME);
						errflg = RES_ERROR;
						goto RETLBL;
					}
					break;
				case DIS_EXP:
					if (Currmark->type != MAR_TAN) {
						fprintf(stderr,"ERROR: transitions %s and %s enabled in marking:\n",pr_trans(aux1,auxtr),pr_trans(aux2,&Arrtrans[Currmark->prow->tr]));
						pr_mark(Currmark,stderr,USE_NAME);
						errflg = RES_ERROR;
						goto RETLBL;
					}
					break;
				}
			}

			if (entrval <= 0) {
				fprintf(stderr,"ERROR: transition %s implicitly inhibited (rate/probability <= 0) in marking:\n",pr_trans(aux1,auxtr));
				pr_mark(Currmark,stderr,USE_NAME);
				fprintf(stderr,"\n");
				errflg = RES_ERROR;
				goto RETLBL;
			}
			/* generate the marking obtained by firing "auxtr" */
			for (i = 0; i < Nplace; i++) {
				input = (auxtr->inp[i] == VARIABLE) ?
					(*(auxtr->vin[i]))(auxtr->i1,auxtr->i2)
					: auxtr->inp[i];
				output = (auxtr->out[i] == VARIABLE) ?
					(*(auxtr->vou[i]))(auxtr->i1,auxtr->i2)
					: auxtr->out[i];
				newmark[i] = Currmark->mk[i] - input + output;
				if (newmark[i] > MAX_TOKEN) {
					fprintf(stderr,"ERROR: token overflow inplace %s: when %s fires\n",pr_place(aux1,&Arrplace[i]),pr_trans(aux2,auxtr));
					pr_mark(Currmark,stderr,USE_NAME);
					exit(ER_CODE);
				}
			}
			/* search (and possibly insert "newmark") */
			auxme = search_ins(&last,newmark);
			/* insert the arc */
			Narcs++;
			auxee = get_EE();
			auxee->row.pt = Currmark->prow;
			auxee->tr = auxtr->num;
			auxee->value = entrval;
			auxee->d_value = d_entrval;
			auxee->col.mk = auxme;
			Currmark->prow = auxee;
			if (IOpt[IOP_DEBUG].val == VAL_YES) {
				fprintf(stderr,"firing transition %s...\n",pr_trans(aux2,auxtr));
			}	
		}
		switch (Currmark->type) {
		case MAR_TAN:	Ntan++;	break;
		case MAR_ABS:	Nabs++;	break;
		case MAR_VAN:	Nvan++;	break;
		}
	}
RETLBL:
	/* simplify graph search in case of error */
	if (errflg == RES_ERROR) {
		Currmark->prow = NULL;
		for (; Currmark != NULL; Currmark = Currmark->pcan) {
			Currmark->prow = NULL;
			Currmark->type = MAR_ABS;
			Nabs++;
		}
	}
	Ntottan = Ntan+Nabs;
	/* classify the markings, compute the number of each type, find loops */
	graph_analysis();
	/* put entries to M0, if required */
	if (FOpt[FOP_ABS_RET_M0].val > 0.0 && Nabs > 0)
		put_ret_m0_entries();
	/* normalize the probabilities */
	normalize_probs();

	/* debugging 
	print_rgraph();
	 debugging */

	return(errflg);
} /* reach_graph() */

ME	**Stack;		/* stack for vanishing */
int	*Lowlink;		/* lowlink value */
int	Store;			/* position in Stack where vanmark are stored */

void	graph_analysis()

/*************************************************************************
It analyzes the reachability graph.
Markings are classified according to their type as
	TAN_M,	ABS_M,	REP_M,	VAN_M,	VLO_M,	A_LO_M.
Indexes are assigned to the markings:
0..(Ntan)-1				for the non absorbing tangible
(Ntan)..(Ntan)+(Nabs)-1			for the absorbing tangible
-1..-(Nvan)				for the vanishing

Ordering for the vanishing markings is defined according to the connected
components: marking in the same connected component are in sequence.
The algorithm used is a variation of Algorithm 5.4, Aho Hopcroft Ullman,
"The Design and Analysis of Computer Algorithms", pag 193.
It calls the recursive function AHUsearch().
*/

{
extern	void	AHUsearch();
	ME	*auxme;
	int	tanind;		/* index of tangible markings */
	int	absind;		/* index of asborbing markings */
	int	vanind;		/* index of vanishing markings */
	int	i;

	Vloop = NULL;
	tanind = 0;
	absind = Ntan;
	/* initialize Store, communicating the value to AHUsearch() */
	Store = Nvan;
	/* allocate the Stack and the Lowlink array */
	Stack = (ME **) new_mem((unsigned int)(sizeof(ME *) * Nvan));
	Lowlink = (int *) new_mem((unsigned int)(sizeof(int) * Nvan));
	Lowlink = &Lowlink[-1];

	/* number the markings, and find the vanishing connected components */
	for (auxme = Pcan; auxme != NULL; auxme = auxme->pcan) {
		if (auxme->index != 0)	/* already numbered */
			continue;
		if (auxme->type == MAR_TAN) {	/* tangible marking */
			auxme->index = tanind++;
			continue;
		}
		if (auxme->type == MAR_ABS) {	/* absorbing marking */
			auxme->index = absind++;
			continue;
		}
		/* otherwise vanishing marking, first of a group of conn. components */
		AHUsearch(auxme);
	}

	/* now all the tangible have been numbered, and all the vanishing are
	stored in Stack, the one to be ordered last is in Stack[0], the one to
	be numbered first is in Stack[Nvan-1]. */

	/* assign vanishing indexes */
	vanind = 0;
	for (i = Nvan; --i >= 0;)
		Stack[i]->index = --vanind;
	free((char *) Stack);
	free((char *) &Lowlink[1]);
} /* graph_analysis */

void	AHUsearch(mark)
ME	*mark;		/* pointer to the ME from which the search starts */

/*************************************************************************
The algorithm used is a variation of Algorithm 5.4, Aho Hopcroft Ullman,
"The Design and Analysis of Computer Algorithms", pag 193.
The considered markings, when removed from the active portion of the
stack, are put in the storage portion of the stack.

A marking,
if considered, and no more in the active part of the stack, has
	index = the vanishing index (a negative number);
if not yet considered has
	index = 0;
if considered but on the stack has
	index = depth first index (a positive number ).
*/

{
	EE	*auxee;
	ME	*auxme;
	LL	*auxll;
	int	absorbing;
	int	firstind,lastind,j,i;
static	int	DFind = 0;	/* deep first index for vanishing markings */
static	int	Vanind = 0;	/* index for the vanishing markings */
static	int	Top = -1;	/* top of the Stack */

	mark->index = ++DFind;
	Lowlink[DFind] = DFind;
	Stack[++Top] = mark;	/* put the marking on the stack */

	/* consider the transitions from this marking */
	for (auxee = mark->prow; auxee != NULL; auxee = auxee->row.pt) {
		auxme = auxee->col.mk;
		if (auxme->type == MAR_TAN || auxme->type == MAR_ABS)
			continue;
		if (auxme->index == 0)	{ /* vanishing, not yet numbered */
			AHUsearch(auxme);
			if (auxme->index > 0 && Lowlink[auxme->index] < Lowlink[mark->index])
				Lowlink[mark->index] = Lowlink[auxme->index];
		} else			/* vanishing, already numbered */
			if (auxme->index < mark->index)
				if (auxme->index > 0) /* auxme is on stack */
					if (auxme->index < Lowlink[mark->index])
						Lowlink[mark->index] = auxme->index;
	}

	/* check if end of a connected component, set correct indexes */
	if (Lowlink[mark->index] == mark->index) {
		/* put connected component in unused part of stack */
		i = 0;
		do {
			i++;
			Stack[--Store] = Stack[Top];
			Stack[Store]->index = --Vanind;
		} while (Stack[Top--] != mark);
		/* the connected component contains i markings */
		auxll = get_LL();
		auxll->length = i;
		auxll->first = Stack[Top + i];
		/* test whether is aloop or vloop */
		absorbing = VAL_YES;
		firstind = Stack[Store]->index;
		lastind = Stack[Store + i - 1]->index;
		for (j = Store + i; absorbing == VAL_YES && j-- != Store;)
			for (auxee = Stack[j]->prow; auxee != NULL; auxee = auxee->row.pt)
				if (auxee->col.mk->index < firstind || auxee->col.mk->index > lastind) {
					absorbing = VAL_NO;
					break;
				}
		if (absorbing == VAL_YES) {
			fprintf(stderr,"ERROR: absorbing vanishing loop\n");
			exit(ER_CODE);
		} else {
			if (i > 1) {	/* true loop */
				for (j = Store + i; j-- != Store;)
					Stack[j]->type = MAR_VLO;
				auxll->pnext = Vloop;
				Vloop = auxll;
			}
		}
	}
} /* AHUsearch() */

int	create_matrix()

/*************************************************************************
It puts the reachability graph in (columnwise) matrix form.
Repeated entries are collapsed.
RETURNS:
RES_ERROR or RES_NOERR.
*/

{
	ME	*pcan;
	EE	*auxee;
	EE	*auxee1;
	int	size,col;
	int	nself;
	int	ndouble;

	nself = 0;
	ndouble = 0;
	Narcs = 0;
	/* the size of Q is tangibles+vanishing+1, index Ntottan is aux  */
	size = Ntottan + Nvan + 1;
	Qrow = (EE **) new_mem((unsigned int)(sizeof(EE *) * size));
	Qcol = (EE **) new_mem((unsigned int)(sizeof(EE *) * size));
	Qrow = &Qrow[Nvan];
	Qcol = &Qcol[Nvan];

	if ( Method_sensitivity == VAL_YES )
	{
	   DerQrow = (EE **) new_mem((unsigned int)(sizeof(EE *) * size));
	   DerQcol = (EE **) new_mem((unsigned int)(sizeof(EE *) * size));
	   DerQrow = &DerQrow[Nvan];
	   DerQcol = &DerQcol[Nvan];
	}

	for (pcan = Pcan; pcan != NULL; pcan = pcan->pcan) {
		/* move the row from pcan into Qcol */
		for (auxee = pcan->prow; auxee != NULL;) {
			if (auxee->col.mk == pcan && pcan->type == MAR_VLO) {
				/* self entry */
				col = Ntottan;
				nself++;
			} else if (Qcol[col = auxee->col.mk->index] != NULL
				&& Qcol[col]->row.ind == pcan->index) {
				/* double entry */
				Qcol[col]->value += auxee->value;
				Qcol[col]->d_value += auxee->d_value;
				col = Ntottan;
				ndouble++;
			} else	/* normal entry */
				Narcs++;
			auxee->col.pt = Qcol[col];
			Qcol[col] = auxee;
			auxee = auxee->row.pt;
			Qcol[col]->row.ind = pcan->index;
		}
		pcan->prow = NULL;
	}
	/* now release the entries pointed by Qcol[Ntottan] */
	if (Qcol[Ntottan] != NULL) {
		rel_EE(Qcol[Ntottan],VAL_TOFROM);
		Qcol[Ntottan] = NULL;
	}
	Direction = VAL_TOFROM;

	if ( Method_sensitivity == VAL_YES )
	{
	   for ( col = -Nvan; col < Ntottan; col++ )
	   {
	      DerQcol[col] = NULL;
	      for ( auxee = Qcol[col]; auxee != NULL; auxee = auxee->col.pt)
	      {
		 if ( auxee->d_value != 0 )
		 {
		   auxee1 = get_EE();
		   auxee1->tr = auxee->tr;
		   auxee1->value = auxee->value;
		   auxee1->d_value = auxee->d_value;
		   auxee1->row.ind = auxee->row.ind;
		   auxee1->col.pt = DerQcol[col];
		   DerQcol[col] = auxee1;
		 }
	      }

	      /* change the ordering to resemble Qcol entries */
	      auxee = DerQcol[col];
	      DerQcol[col] = NULL;

	      while ( auxee != NULL )
	      {
		 auxee1 = auxee;
		 auxee = auxee->col.pt;
		 auxee1->col.pt = DerQcol[col];
		 DerQcol[col] = auxee1;
	      }
	   }
	   DerDirection = VAL_TOFROM;
	}

	fprintf(stderr,"After elimination of redundant arcs:\n");
	if (nself > 0)
		fprintf(stderr,"\t# of self-arcs eliminated:\t%6d\n",nself);
	if (ndouble > 0)
		fprintf(stderr,"\t# of double-arcs compacted:\t%6d\n",ndouble);
	if (FOpt[FOP_ABS_RET_M0].val > 0.0 && Nabs > 0)
		fprintf(stderr,"\t# of arcs added toward M0:\t%6d\n",Nabs);
	fprintf(stderr,"\t# of remaining arcs:\t\t%6d\n",Narcs);
	Nentries = Narcs;
	return(RES_NOERR);
} /* create_matrix() */

void print_rgraph()
{
   ME *pcan;
   EE *auxee;

   fprintf(stderr,"Printing the reachability graph\n");
   for ( pcan = Pcan; pcan != NULL; pcan = pcan->pcan)
   {
      fprintf(stderr," Marking Index: %d\n", pcan->index );
      for( auxee = pcan->prow; auxee != NULL; auxee= auxee->row.pt)
      {
	 fprintf(stderr,"transition: %d, value: %.12g, dvalue: %.12g\n",
		  auxee->tr, auxee->value, auxee->d_value );
      }
   }
}

void print_qcol()
/* debugging function. Prints Qcol matrix */
{
   EE *auxee;
   int index;

   fprintf(stdout,"The QCOl Matrix: \n");
   for( index = -Nvan; index < Ntottan; index++ )
   {
      fprintf(stdout,"%d   ", index );
      for( auxee = Qcol[index]; auxee != NULL; auxee = auxee->col.pt)
      {
	 fprintf(stdout,"%d:%.8f  ", auxee->row.ind, auxee->value );
      }
      fprintf(stdout,"\n");
   }

}

void print_qrow()
/* debugging function. Prints Qrow matrix */
{
   EE *auxee;
   int index;

   fprintf(stdout,"The QROW Matrix: \n");
   for( index = -Nvan; index < Ntottan; index++ )
   {
      fprintf(stdout,"%d   ", index );
      for( auxee = Qrow[index]; auxee != NULL; auxee = auxee->row.pt)
      {
	 fprintf(stdout,"%d:%.8f  ", auxee->col.ind, auxee->value );
      }
      fprintf(stdout,"\n");
   }

}

void print_derqcol()
/* debugging function. Prints Qcol matrix */
{
   EE *auxee;
   int index;

   if ( Method_sensitivity == VAL_NO )
      return;

   fprintf(stdout,"The DERQCOl Matrix: \n");
   for( index = -Nvan; index < Ntottan; index++ )
   {
      fprintf(stdout,"%d   ", index );
      for( auxee = DerQcol[index]; auxee != NULL; auxee = auxee->col.pt)
      {
	 fprintf(stdout,"%d:%.8f  ", auxee->row.ind, auxee->d_value );
      }
      fprintf(stdout,"\n");
   }

}

void print_derqrow()
/* debugging function. Prints Qrow matrix */
{
   EE *auxee;
   int index;

   if ( Method_sensitivity == VAL_NO )
      return;

   fprintf(stdout,"The DERQROW Matrix: \n");
   for( index = -Nvan; index < Ntottan; index++ )
   {
      fprintf(stdout,"%d   ", index );
      for( auxee = DerQrow[index]; auxee != NULL; auxee = auxee->row.pt)
      {
	 fprintf(stdout,"%d:%.8f  ", auxee->col.ind, auxee->d_value );
      }
      fprintf(stdout,"\n");
   }

}
